//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, C++ rewrite 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  commands.cpp
//! \brief Function fridaCommand: executes one main-level command.

#include "defs.hpp"

#include <cstring>
#include <cstdio>
#include <cmath>

#include "../trivia/file_ops.hpp"
#include "../readplus/ask.hpp"

#include "config.hpp"
#include "olf.hpp"
#include "mem.hpp"
#include "edif.hpp"
#include "import.hpp"
#include "func.hpp"
#include "fregistry.hpp"
#include "geni.hpp"
#include "dualplot.hpp"
#include "opr.hpp"
#include "plot.hpp"
#include "file_in.hpp"
#include "file_out.hpp"
#include "rssm.hpp"
#include "manip.hpp"
#include "curve.hpp"
#include "fit.hpp"
#include "expr.hpp"
#include "obj.hpp"
#include "integrate.hpp"
#include "special.hpp"


//! Executes cmd if is a Frida command, and returns true when successful.

bool frida_command( string cmd )
{
    int i;

    // Extract overwrite marker:
    NOlm::overwrite = 0;
    if ( cmd.size()>1 && cmd[cmd.size()-1]=='!' ) {
        NOlm::overwrite = 1;
        cmd.erase(cmd.size()-1,1);
    }

    // If there are new files, make them new default
    NOlm::sel_collect_begin();

    if (cmd == "h" || cmd == "?") {
        cout <<
            "This is Frida, version " VERSION "\n"
            "Plain commands:\n"
            "  a    add to plot\n"
            "  p    plot (g for list of other plot commands)\n"
            "  t    show execution time (for development)\n"
            "  !    execute shell command\n"
            "  quit terminate Frida session\n"
            "Command groups (type initial letter for help):\n"
            "  b    backup of online files\n"
            "  c    curves and fits\n"
            "  d    directory (inspect internal files)\n"
            "  e    edit (change textual parameters)\n"
            "  g    graphics setup\n"
            "  f    file input, output, creation\n"
            "  m    manipulate data (select, bin, reorganise, ...)\n"
            "  o    operate on variables, compute integrals, ...\n"
            "  r    read raw data (currently only for SPHERES)\n"
            "  _    special / temporary commands\n"
            "Modifier (appended to command):\n"
            "  !    operate in place (overwrite)\n"
            "More help:\n"
            "  hc   constants\n"
            "  hf   functions of scalar arguments\n"
            "  hid  dimension-reducing functionals of data (replaced 'oio')\n"
            "  hic  dimension-reducing functionals of curves\n"
            "  ho   operators by precedence\n"
            "  hr   handling resolution functions (convolution, Dirac delta function)\n"
            "  hv   variables\n"
            "  \\h   meta commands and macros\n"
            ;

    } else if (cmd == "hc") {
        cout <<
            "Constants for use in expressions\n"
            "  e   2.718.. Euler number\n"
            "  pi  3.141.. \n"
            "  inf infinity\n"
            ;
    } else if (cmd == "hf") {
        GFuncRegistry->display_functions();
    } else if (cmd == "hi") {
        cout <<
            "  hid  dimension-reducing functionals of data (replacing 'oio')\n"
            "  hic  dimension-reducing functionals of curves\n"
            ;
    } else if (cmd == "hid" || cmd == "oio") {
        NGeni::display_genis();
    } else if (cmd == "hic") {
        NCvin::display_cvins();
    } else if (cmd == "ho") {
        GFuncRegistry->display_operators();
    } else if (cmd == "hr") {
        cout <<
            "Generic form of fit function that is to be convolved with a resolution spectrum:\n"
            "    p0+p1*resol(p9)+p2*conv(f(t,p3,...,p8),p9)\n"
            "Notes:\n"
            "    - flat background of amplitude p0 does not need convolution\n"
            "    - 'resol' copies the resolution data\n"
            "      (implicit convolution of a Dirac delta distribution with the resolution)\n"
            "    - 'conv' convolutes the given function f(t,...) with the resolution\n"
            "    - p9 is a shift in t; most often it can be set to zero\n"
            "      (a zero shift argument may be omitted from 'conv', but not from 'resol')\n"
            "To activate the convolution:\n"
            "    - use command 'cv' to provide resolution file reference\n"
            "If function f has a sharp peak (strong variation within one resolution bin):\n"
            "    - use 'pconv(F(t,...))', where F is the primitive of f\n"
            ;
    } else if (cmd == "hv") {
        cout <<
            "Variables for use in expressions\n"
            "  t         dummy argument in curve definition\n"
            "  k         index running over files in online memory\n"
            "  j         index running over slices (spectra or curves) in a file\n"
            "  i         index running over points in a spectrum\n"
            "  nk        number of files in online memory\n"
            "  nj[K]     number of slices (spectra or curves) in a file\n"
            "  ni[K,J]   number of points in a spectrum\n"
            "  rL[K]     file-wide parameter number L\n"
            "  zL[K,J]   slice-wide parameter number L\n"
            "  pL[K,J]   curve parameter number L\n"
            "  f[K,J](A) curve's function value for argument A\n"
            "  x[K,J,I]  x-value of a point in a spectrum\n"
            "  y[K,J,I]  y-value of a point in a spectrum\n"
            "  dy[K,J,I] error of y[K,J,I]\n"
            "Note on index placeholders K,J,I:\n"
            "  - each of them can be omitted\n"
            "  - when omitted, the respective default value k, j, or i is used\n"
            "  - no comma needed when only omitted indices follow\n"
            "  - the bracket may also be omitted altogether\n"
            "  - hence are equivalent: x = x[k,j,i] = x[k,j,] = x[k,j] = x[k,] = x[k]\n"
            ;

    } else if (cmd == "b") {
        cout << "Backup commands:\n"
            "  bc   copy to online backup (automatically before cf)\n"
            "  br   retrieved from online backup\n"
            ;

    } else if (cmd == "bc") {
        NOlm::mem_backup_store();

    } else if (cmd == "br") {
        NOlm::mem_backup_retrieve();

    } else if (cmd == "c") {
        cout << "Curve and fit commands:\n"
            "  cc   create\n"
            "  cca  create artificial curve (no data reference)\n"
            "Basic operations:\n"
            "  cm   modify function\n"
            "  cd   set data file to fit\n"
            "  cv   set file to convolute with (cv-: no convolution)\n"
            "  cr   set fit range restriction\n"
            "  cf   fit\n"
            "  cfs  fit, allow slow convolution\n"
            "  cp   print parameters\n"
            "Parameter attributes:\n"
            "  cx   fix parameters\n"
            "  cg   make parameters global (same for all curves in the file)\n"
            "  cu   unfix parameters (remove fixed and global attributes)\n"
            "File attributes:\n"
            "  cwc  weighing: constant\n"
            "  cwl  weighing: logarithmic\n"
            "  cwv  weighing: reciprocal variance (default)\n"
            "  cwvd weighing: reciprocal variance, data only\n"
            "  cwvc weighing: reciprocal variance, curve only\n"
            "  cgd+ graphics: take x points from data reference file\n"
            "  cgd- graphics: create x points by dividing the plot window\n"
            "Frida-wide settings:\n"
            "  cs?  show fit parameters, and indicate commands to modify them\n"
            "  cn?  show parameters for numeric integration, and indicate commands to modify them\n"
            ;

    } else if (cmd == "ci") {
        cout << "obsolete command: 'ci 1' has become 'oi p1'\n";

    } else if (cmd == "cc" ) {
        NCurveFile::create_fitcurve();
    } else if (cmd == "cca" ) {
        NCurveFile::create_freecurve();

    } else if (cmd == "cm" ) {
        NCurveFile::change_expr();

    } else if (cmd == "cd" || cmd == "cv" || cmd == "cv-" ) {
        NCurveFile::set_file_reference( cmd.substr(1) );

    } else if (cmd == "cr" ) {
        NCurveFile::change_range();

    } else if (cmd == "cp") {
        NCurveFile::show_pars();

    } else if (cmd == "cf") {
        NCurveFit::fit( false );
    } else if (cmd == "cfs") {
        NCurveFit::fit( true );
    } else if (cmd == "cx" || cmd == "cg" || cmd == "cu" ) {
        NCurveFile::set_par_attr( cmd[1] );
    } else if (cmd.substr(0,2) == "cw" || cmd.substr(0,2) == "cg") {
        NCurveFile::set_properties( cmd.substr(1) );

    } else if (cmd.substr(0,2) == "cs") {
        NCurveFit::set_fit_tuning_pars(cmd.substr(2));

    } else if (cmd.substr(0,2) == "cn") {
        NCvin::set_integration_tuning_pars(cmd.substr(2));


    } else if ( cmd == "d" ) {
        cout << "Directory commands: inspect internal files\n"
            "  df   list of files\n"
            "  dz   z-values and spectral ranges\n"
            "  dp   x|y-values\n"
            "  dy   y-values\n"
            "  dr   real parameters\n"
            "  dc   coordinate names and units\n"
            "  dd   doc lines\n"
            ;

    } else if (cmd == "df") {
        NEdif::show_files();
    } else if (cmd == "dz") {
        NEdif::show_spectra();
    } else if (cmd == "dp" ) {
        NOperate::show( "xyd" );
    } else if (cmd == "dy" ) {
        NOperate::show( "y" );
    } else if (cmd == "dr") {
        NEdif::show_numpars();
    } else if (cmd == "dc") {
        NEdif::show_coord();
    } else if (cmd == "dd") {
        NEdif::show_doc();

    } else if ( cmd == "e" ) {
        cout << "Edit commands:\n"
            "  ef         file name\n"
            "  ec<coord>  name and unit of <coord>=x,y,z0,..\n"
            "  ed         documentation lines\n"
            ;

    } else if (cmd.substr(0,2) == "ec") {
        NEdif::edit_coord( cmd.substr(2) );
    } else if (cmd == "ed") {
        NEdif::edit_doc();
    } else if (cmd == "ef") {
        NEdif::edit_fnam();


    } else if (cmd == "f") {
        cout << "File commands:\n"
            "  fl   load from disk\n"
            "  fs   save to disk\n"
            "  fso  save to disk, allow overwrite\n"
            "  fe0  export to old y08 format (obsolete!)\n"
            "  fe1  export (only z and y, one line per spectrum)\n"
            "  fe2  export (one block per spectrum: z header and x-y lines)\n"
            "  fc   copy\n"
            "  fx   exchange (copy, then delete source)\n"
            "  fdel delete\n"
            "  fm   make\n"
            "  fr   read in\n"
            "  ft(h|v|c)[m][s][d]  read tab\n"
            "       h horizontal (only y values, one line per spectrum)\n"
            "       v vertical (one x-y pair per line)\n"
            "       c choose vertical columns\n"
            "       m multi-block (blocks begin with z line)\n"
            "       s from script\n"
            "       d input includes dy\n"
            ;

    } else if (cmd == "fl") {
        NFileIn::load();
    } else if (cmd == "fs") {
        NFileOut::save( "yda" );
    } else if (cmd == "fso") {
        NFileOut::save( "yda", true );
    } else if (cmd == "fe0") {
        NFileOut::save( "y08" );
    } else if (cmd == "fe0o") {
        NFileOut::save( "y08", true );
    } else if (cmd == "fe1") {
        NFileOut::save( "csv" );
    } else if (cmd == "fe2") {
        NFileOut::save( "tab" );
    } else if (cmd == "fc") {
        NOlm::mem_copy();
    } else if (cmd == "fx") {
        NOlm::mem_xchange();
    } else if (cmd == "fdel") {
        NOlm::mem_del();
    } else if (cmd == "fm") {
        NImport::make_grid();
    } else if (cmd == "fr") {
        NImport::read_in();
    } else if (cmd.substr(0,2) == "ft") {
        NImport::read_tab( cmd.substr(2) );


    } else if (cmd == "g") {
        cout <<
            "Graphics commands:\n"
            "  p    plot\n"
            "  pr   plot with new autoranges\n"
            "  pn   plot next spectrum\n"
            "  pv   plot previous spectrum\n"
            "  a    add to plot\n"
            "  an   add next spectrum\n"
            "  av   add previous spectrum\n"
            "  gd   graphic dialog (feed commands to gnuplot)\n"
            "Save for print:\n"
            "  gp   plot to complete .ps file\n"
            "  gf   plot to short .psX file\n"
            "Change window:\n"
            "  gw   list of plot windows\n"
            "  g<n> switch to plot window <n>\n"
            "Set axes:\n"
            "  gx   set x range\n"
            "  gy   set x range\n"
            "  ga   auto-ranges\n"
            "  gxa  x auto-range\n"
            "  gya  y auto-range\n"
            "  gxl  log x (toggle; or use gxl+ gxl-)\n"
            "  gyl  log y\n"
            "  gxf  force x into frame (toggle; or use gxf+ gxf-)\n"
            "  gyf  force y into frame (toggle; or use gyf+ gyf-)\n"
            "Other setup:\n"
            "  ge   plot error bars (toggle, or ge+ ge-)\n"
            "  gnd  set maximum number of data points\n"
            "  gnc  number of equidistant base points for curve plotting\n"
            "  gcr  allow iterative refinement for curve plotting?\n"
            "Help:\n"
            "  gsym explain plot symbols\n"
            ;

    } else if ( cmd == "gx" ) {
        NPloWin::Plots[NPloWin::iPlot]->X.ask_and_set( "Plot x range" );
    } else if ( cmd == "gy" ) {
        NPloWin::Plots[NPloWin::iPlot]->Y.ask_and_set( "Plot y range" );
    } else if ( cmd == "ga" ) {
        NPloWin::Plots[NPloWin::iPlot]->X.set_auto();
        NPloWin::Plots[NPloWin::iPlot]->Y.set_auto();
    } else if ( cmd == "gxa" ) {
        NPloWin::Plots[NPloWin::iPlot]->X.set_auto();
    } else if ( cmd == "gya" ) {
        NPloWin::Plots[NPloWin::iPlot]->Y.set_auto();
    } else if ( cmd.substr(0,2) == "gx" ||
                cmd.substr(0,2) == "gy" ||
                cmd.substr(0,2) == "ge" ) {
        NPloWin::Plots[NPloWin::iPlot]->set_aux( cmd );
    } else if ( cmd=="gnd" ) {
        NPloWin::Plots[NPloWin::iPlot]->maxpoints =
            iask( "Maximum number of data points in plot",
                  NPloWin::Plots[NPloWin::iPlot]->maxpoints );
    } else if ( cmd=="gnc" ) {
        NPloWin::Plots[NPloWin::iPlot]->equipoints =
            iask( "Number of equidistant base points for curve plotting",
                  NPloWin::Plots[NPloWin::iPlot]->equipoints );
    } else if ( cmd=="gcr" ) {
        NPloWin::Plots[NPloWin::iPlot]->refine =
            bask( "Allow iterative refinement for curve plotting",
                  NPloWin::Plots[NPloWin::iPlot]->refine );
    } else if (cmd == "gp" || cmd == "gf" ) {
        string ps_outdir = CNode::eval( "psdir" )->to_s();
        string ps_head = CNode::eval( "pshead" )->to_s();
        string ps_dict = cmd=="gp" ? CNode::eval( "psdict" )->to_s() : "";
        NPloWin::Plots[NPloWin::iPlot]->write_postscript( ps_outdir, ps_head, ps_dict );
    } else if ( cmd == "gw" ) {
        for( int i=0; i<NPloWin::nPlot; ++i ){
            printf( " %c %1i %s\n", i==NPloWin::iPlot ? '*' : ' ', i,
                    NPloWin::Plots[i]->info().c_str() );
        }
    } else if ( sscanf( cmd.c_str(), "g%i", &i )==1 ) {
        if( i<NPloWin::nPlot )
            NPloWin::iPlot = i;
        else if ( i==NPloWin::nPlot ){
            NPloWin::iPlot = NPloWin::nPlot;
            NPloWin::Plots.push_back( new CPlot( NPloWin::nPlot++, false, false ) );
        } else
            throw S("invalid graphic window number");
    } else if (cmd == "gd") { // graph dialog
        string cmd;
        cout << "entering mygnuplot - to leave, type q\n";
        while(1) {
            cmd = sask("mygnuplot> ");
            if (cmd.substr(0,1)=="q")
                break;
            NPloWin::Plots[NPloWin::iPlot]->gp_write(cmd);
        }
    } else if (cmd == "gsym") {
        system( ("gv " + CNode::eval( "psgsym" )->to_s() + "&").c_str() );
    } else if (cmd == "p") {
        NPlot::plot( NPloWin::Plots[NPloWin::iPlot], false );
    } else if (cmd == "pr") {
        NPloWin::Plots[NPloWin::iPlot]->X.set_auto();
        NPloWin::Plots[NPloWin::iPlot]->Y.set_auto();
        NPlot::plot( NPloWin::Plots[NPloWin::iPlot], false );
    } else if (cmd == "pn") {
        NPlot::plot( NPloWin::Plots[NPloWin::iPlot], false, "next" );
    } else if (cmd == "pv") {
        NPlot::plot( NPloWin::Plots[NPloWin::iPlot], false, "prev" );
    } else if (cmd == "a") {
        NPlot::plot( NPloWin::Plots[NPloWin::iPlot], true );
    } else if (cmd == "an") {
        NPlot::plot( NPloWin::Plots[NPloWin::iPlot], true, "next" );
    } else if (cmd == "av") {
        NPlot::plot( NPloWin::Plots[NPloWin::iPlot], true, "prev" );


    } else if (cmd == "k") {
        cout << "Commands to modify the internal file selection:\n"
            "  kn    next\n"
            "  kv    previous\n";
    } else if (cmd == "kn") {
        NOlm::sel_increment( +1 );
    } else if (cmd == "kv") {
        NOlm::sel_increment( -1 );

    } else if (cmd == "m") {
        cout << "Commands to manipulate data:\n"
            "  m/    protect spectra from forthcoming operations\n"
            "  m/-   lift protections\n"
            "Generic manipulations (expression driven):\n"
            "  md    generic delete\n"
            "  mr    generic retain\n"
            "Per point:\n"
            "  mpd   delete\n"
            "  mpr   retain\n"
            "  mpa   average (=bin), list controlled\n"
            "  mpaf  average, reduce number of points by fixed factor\n"
            "  mpaea average according to error bound (absolute)\n"
            "  mpaer average according to error bound (relative)\n"
            "  mpo   sort\n"
            "  mpq   average when x equal\n"
            "  mpsym symmetrize: map -x -> x, then average\n"
            "  mpe-  remove error bar\n"
            "Per spectrum:\n"
            "  msd   delete\n"
            "  msr   retain\n"
            "  msa   average\n"
            "  msb   break into files\n"
            "  msj   join\n"
            "  ms*   spawn spectra\n"
            "  msx   z_max <-> x\n"
            "  mso   sort by expression\n"
            "  msoz  sort by z-values\n"
            "  mzx   exchange / rotate z coordinates\n"
            "  mz-   delete z coordinate\n"
            "Per file:\n"
            "  mfj   merge (\"join\")\n"
            "  mfj+  merge, adding z=number-of-input-file\n"
            "  mfjp  merge pointwise\n"
            "  mr-   remove real par\n"
            "Interpolation operations:\n"
            "  mi    locate y in x' to determine y'(x'=y)\n"
            "Histogram operations:\n"
            "  mhm   bin data\n"
            ;

    } else if (cmd == "m/") {
        NManip::freeze_slices();
    } else if (cmd == "m/-") {
        NManip::unfreeze_all_slices();
    } else if (cmd == "md") {
        NOperate::select(1);
    } else if (cmd == "mr") {
        NOperate::select(0);
    } else if (cmd == "mpd") {
        NManip::points_select( true );
    } else if (cmd == "mpr") {
        NManip::points_select( false );
    } else if (cmd == "mpa") {
        NManip::points_rebin();
    } else if (cmd == "mpaf") {
        NManip::points_rebin_by_factor();
    } else if (cmd.substr(0,4) == "mpae") {
        NManip::points_rebin_by_err( cmd.substr(4) );
    } else if (cmd == "mpo") {
        NManip::points_sort();
    } else if (cmd == "mpq") {
        NManip::points_rebin_duplicates();
    } else if (cmd == "mpsym") {
        NManip::points_symmetrize();
    } else if (cmd == "mpe-") {
        NManip::points_remove_err();
    } else if (cmd == "msd") {
        NManip::slices_select( true );
    } else if (cmd == "msr") {
        NManip::slices_select( false );
    } else if (cmd == "msa") {
        NManip::slices_rebin();
    } else if (cmd == "msb") {
        NManip::slices_break();
    } else if (cmd == "msj") {
        NManip::slices_merge();
    } else if (cmd == "ms*") {
        NManip::slices_spawn();
    } else if (cmd == "msx") {
        NManip::exchange_x_z();
    } else if (cmd == "mzx") {
        NManip::zcoords_reorder();
    } else if (cmd == "mz-") {
        NEdif::show_coordZ();
        NManip::zcoord_delete();
    } else if (cmd == "mso") {
        NManip::slices_sort();
    } else if (cmd == "msoz") {
        NManip::slices_sort_by_z();
    } else if (cmd == "mfjp") {
        NManip::files_merge_pointwise();
    } else if (cmd.substr(0,3) == "mfj") {
        NManip::files_merge( cmd.substr(3) );
    } else if (cmd == "mr-") {
        NEdif::del_numpar();
    } else if (cmd == "mi") {
        NManip::interpolate();
    } else if (cmd == "mhm") {
        NManip::make_histogram();


    } else if (cmd == "o") {
        cout << "Commands to operate on variables or spectra:\n"
            "  ox       on x\n"
            "  oy       on y\n"
            "  oz0,..   on z0,..\n"
            "  oz+      add new z value\n"
            "  or0,..   on r0,..\n"
            "  or+      add new real par\n"
            "  op0,..   on p0,.. (fit parameter)\n"
            "  op*      on all fit parameters\n"
            "  opa      take absolute value\n"
            "  opc      copy from another curve\n"
            "  ofi      functional integrate\n"
            "  ofs      summation up to current index\n"
            "  ofd      functional derivate\n"
            "  ofac     functional autocorrelate\n"
            "  oi       rank-reducing operations (""generalized integrals"")\n"
            "  oixy     choose two columns, save them as x and y\n"
            "  oixyd    choose three columns, save them as x, y, dy\n"
            ;

    } else if (cmd == "op") {
        throw S("use op<n> to modify one parameter, or op* to modify all");
    } else if (cmd == "op*") {
        NCurveFile::pars_query();
    } else if (cmd == "opa") {
        NCurveFile::pars_abs();
    } else if (cmd == "opc") {
        NCurveFile::pars_copy();
    } else if (cmd[0]=='o' && strchr("xydzpr", cmd[1])) {
        NOperate::Pointwise( cmd.substr(1) );
    } else if (cmd.substr(0,2) == "of") {
        NOperate::Functional( cmd.substr(2) );
    } else if (cmd == "oi") {
        NOperate::Integral();
    } else if (cmd.substr(0,4) == "oixy") {
        NOperate::IntXY( cmd.substr(2) );


    } else if (cmd == "r") {
        cout << "Read raw data from SPHERES:\n"
            "  ry   load energy spectra\n"
            "  ry2  load also spectra for open state\n"
            "  ry4  load also chopper histograms\n"
            "  ry8  load spectra for two directions separately\n"
            "  ryd  load only spectra for two directions\n"
            "  ryo  load only spectra for open state\n"
            "  ryc  raw counts (no normalization)\n"
            "  rs   load series of energy spectra\n"
            "  rs2  load also spectra for open state\n"
            ;

    } else if (cmd == "ry" ) {
        NRSSM::read_spec( 0 );
    } else if (cmd == "ry2") {
        NRSSM::read_spec( 1 );
    } else if (cmd == "ry4") {
        NRSSM::read_spec( 3 );
    } else if (cmd == "ry8") {
        NRSSM::read_spec( 7 );
    } else if (cmd == "ryd") {
        NRSSM::read_spec( 12 );
    } else if (cmd == "ryo") {
        NRSSM::read_spec( 9 );
    } else if (cmd == "ryc") {
        NRSSM::read_spec( 16 );
    } else if (cmd == "rs") {
        NRSSM::read_series( 0 );
    } else if (cmd == "rs2") {
        NRSSM::read_series( 1 );


    } else if (cmd == "!") {
        string sh_cmd;
        sh_cmd = sask("shell command");
        triv::system( sh_cmd );


    } else if (cmd == "_") {
        cout << "Special programs:\n"
            "  _fc      cosine Fourier transform\n"
            "  _t       test (may change at any moment)\n"
            ;


    } else if (cmd == "_fc") {
        NSpecial::FourierCosine();

    } else if (cmd == "_t") {
        NSpecial::Test();

    } else if ( cmd == "qui" || cmd=="quit" ) {
        exit(0);

    } else {
        return false; // command not found
    }

    // If there are new files, make them new default
    NOlm::sel_collect_end();

    return true;
}
