//***************************************************************************//
//* libtrivia: various little routines for C++                              *//
//* vector_ops: vector operations                                           *//
//* (C) Joachim Wuttke 2001, 2012-                                          *//
//* http://apps.jcns.fz-juelich.de                                          *//
//***************************************************************************//

#include <cmath>
#include <string>
#include <vector>
#include <algorithm>
#include <numeric>

using namespace std;

#include "vector_ops.hpp"
#include "string_convs.hpp"


//***************************************************************************//
//*  vector operations                                                      *//
//***************************************************************************//

//! Inserts val into vector *V, presuming the latter is sorted.

void triv::insert_in_sorted( vector<double> *V, double val )
{
    for (size_t i=0; i<V->size(); ++i) {
        if (val>(*V)[i]) continue;
        if (val==(*V)[i]) return;
        V->insert(V->begin()+i, val);
        return;
    }
    V->push_back(val);
}

//! Eliminates double entries from vector V.

void triv::make_unique( vector<double> *V, double tolabs, double tolrel )
{
    // tolrel z.Zt. nicht benutzt
    vector<double> aux;
    double diff;
    if (!V->size()) return;
    aux.push_back((*V)[0]);
    for (size_t ii=1; ii<V->size(); ++ii)
        if((diff=fabs((*V)[ii]-aux.back()))>tolabs)
            aux.push_back((*V)[ii]);
    *V = aux;
}

//! Returns rue unless V[i]<=V[i-1] for some i.

bool triv::is_ascending( const vector<double>& V )
{
    for( size_t i=2; i<V.size(); ++i )
        if ( V[i] <= V[i-1] )
            return false;
    return true;
}

//! Returns true if grid is equidistant, sets *step.

bool triv::is_equidist( double *step, const vector<double>& V )
{
    size_t n = V.size();
    if( n < 2 )
        return false;  // vector too short to determine equidistance
    *step = ( V[n-1] - V[0] ) / (n-1);
    for ( size_t i=1; i<n-1; ++i )
        if( fabs( V[i] - (V[0]+i*(*step)) ) > 1e-12*fabs(*step)+1e-200 )
            return false;
    return true;
}

//! Returns array of indices sorted according to V.

vector<size_t> triv::sorted_indices(vector<double> const& V)
{
    vector<size_t> I(V.size());
    // C++11, not yet supported by gcc-4.4:
    // iota(begin(I), end(I), 0);
    // workaround:
    for( size_t i=0; i<I.size(); ++i )
	I[i] = i;
    // end workaround.
    sort( begin(I), end(I), [&](size_t a, size_t b) { return V[a] < V[b]; } );
    return I;
}


//! Increments all members, provided they remain in the range 0...siz.

void triv::increment_indices( vector<int>& v, int incr, int siz )
{
    vector<int> ret;
    for( int k: v ){
        int kout = k+incr;
        if ( kout<0 || kout>=siz )
            return; // leave v unchanged
        ret.push_back( kout );
    }
    v = ret;
}

//! Returns string representation of index vector, making use of ranges where possible.

string triv::indices_to_s(const vector<int>& v)
{
    string ret;
    int n = v.size();
    if( !n )
        return "";
    int iout = 0; // next to be written
    while( true ) {
        if ( iout+2<n ) {
            int step = v[iout+1]-v[iout];
            int i=iout+1;
            for ( ; i<n-1; ++i )
                if( v[i+1]-v[i]!=step )
                    break;
            if( i>iout+1 ) {
                ret += strg(v[iout]) + ":" + strg(v[i]);
                if( step!=1 )
                    ret += ":" + strg(step);
                iout = i+1;
            } else {
                ret += strg(v[iout]);
                iout++;
            }
        } else {
            ret += strg(v[iout]);
            iout++;
        }
        if ( iout>=n )
            break;
        ret += ",";
    }
    return ret;
}
